*/
    QuickMSG
    Copyright (C) 2014  Jeroen Vreeken <jeroen@vreeken.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#include "gpg.h"
#include "settings.h"

#include <stdlib.h>
#include <string.h>
#include <gpgme.h>
#include <unistd.h>

struct gpg {
	gpgme_ctx_t ctx;
};

static gpgme_error_t passphrase_cb(void *hook, const char *uid_hint, 
    const char *passphrase_info, int prev_was_bad, int fd)
{
	char *pass = settings_gpg_pass_get();
	
	printf("pass cb\n");
	if (prev_was_bad)
		return GPG_ERR_CANCELED;
	
	write(fd, pass, strlen(pass));
	write(fd, "\n", 1);

	printf("Send pass\n");
	
	return 0;
}

struct gpg *gpg_alloc(void)
{
	struct gpg *gpg;
	gpgme_error_t err;
	gpgme_engine_info_t enginfo;
	
	gpg = calloc(1, sizeof(struct gpg));
	
	gpgme_check_version(NULL);

	/* get engine information */
	err = gpgme_get_engine_info(&enginfo);
	if(err != GPG_ERR_NO_ERROR) {
		printf("Error getting engine info\n");
		goto err;
	}
	
	err = gpgme_new(&gpg->ctx);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating gpg context\n");
		goto err;
	}

	/* set protocol to use in our context */
	err = gpgme_set_protocol(gpg->ctx,GPGME_PROTOCOL_OpenPGP);
	if(err != GPG_ERR_NO_ERROR) {
		printf("Error setting protocol");
		goto err;
	}

	/* set engine info in our context; I changed it for ceof like this:
	err = gpgme_ctx_set_engine_info (ceofcontext, GPGME_PROTOCOL_OpenPGP,
               "/usr/bin/gpg","/home/user/nico/.ceof/gpg/");

     	 but I'll use standard values for this example: */

	err = gpgme_ctx_set_engine_info (gpg->ctx, GPGME_PROTOCOL_OpenPGP,
               enginfo->file_name,enginfo->home_dir);
	if(err != GPG_ERR_NO_ERROR) {
		printf("Error setting engine info\n");
		goto err;
	}
	
	gpgme_set_armor (gpg->ctx, 1);
	gpgme_set_passphrase_cb(gpg->ctx, passphrase_cb, NULL);
	
	return gpg;

err:
	free(gpg);
		
	return NULL;
}

void gpg_free(struct gpg *gpg)
{
	free(gpg);
}

int gpg_add_key(struct gpg *gpg, struct attachment *a)
{
	gpgme_data_t dh;
	gpgme_error_t err;
	int ret = 0;

	err = gpgme_data_new_from_mem(&dh, a->data, a->length, 0);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating data struct\n");
		ret = -1;
		goto err_no_dh;
	}
	
	err = gpgme_op_import (gpg->ctx, dh);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error importing key\n");
		ret = -1;
	}
	
	gpgme_data_release(dh);
err_no_dh:	
	return ret;
}

struct attachment *gpg_get_key(struct gpg *gpg, char *address)
{
	gpgme_error_t err;
	gpgme_data_t dh;
	struct attachment *a = NULL;
	void *data;
	size_t length;
	
	err = gpgme_data_new(&dh);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating data struct\n");
		goto err_no_dh;
	}

	err = gpgme_op_export (gpg->ctx, address, 0, dh);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating data struct\n");
		goto err;
	}

	a = attachment_alloc();
	if (!a)
		goto err;

	attachment_type_set(a, "pgp-keys");
	attachment_basetype_set(a, "application");
	
	data = gpgme_data_release_and_get_mem (dh, &length);
	
	attachment_data_set(a, data, length);
	
	gpgme_free(data);
	return a;
err:
	gpgme_data_release(dh);
	return a;

err_no_dh:	
	return NULL;
}

char *gpg_get_fingerprint(struct gpg *gpg, char *address)
{
	gpgme_error_t err;
	char *fp = NULL;
	
	err = gpgme_op_keylist_start (gpg->ctx, address, 0);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating data struct\n");
		goto err_no_fp;
	}
	
	do {
		gpgme_key_t key;
		gpgme_subkey_t subkey;

		err = gpgme_op_keylist_next (gpg->ctx, &key);

		subkey = key->subkeys;
		if (!subkey)
			continue;
		
		fp = strdup(subkey->fpr);

		gpgme_key_release(key);
	} while (err == GPG_ERR_EOF);

	return fp;
err_no_fp:
	return NULL;
}

struct attachment *gpg_decrypt_verify(struct gpg *gpg, struct attachment *a, char *from)
{
	gpgme_error_t err;
	gpgme_data_t dh_e, dh_p = NULL;
	struct attachment *plain = NULL;
	void *data;
	size_t length;
	gpgme_verify_result_t vr;
	char *fp;

	fp = gpg_get_fingerprint(gpg, from);
	if (!fp) {
		printf("Could not find sender fingerprint\n");
		return NULL;
	}
	
	err = gpgme_data_new_from_mem(&dh_e, a->data, a->length, 0);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating encrypted data struct\n");
		goto err_no_dh;
	}

	err = gpgme_data_new(&dh_p);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating plain data struct\n");
		goto err_dh_e;
	}

	err = gpgme_op_decrypt_verify (gpg->ctx, dh_e, dh_p);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error decoding data\n");
		goto err_dh_p;
	}

	vr = gpgme_op_verify_result(gpg->ctx);
	if (strcmp(fp, vr->signatures->fpr)) {
		printf("Fingerprint does not match: %s != %s\n",
		    vr->signatures->fpr, fp);
		goto err_dh_p;
	}
 
	plain = attachment_alloc();
	if (!plain)
		goto err_dh_p;

	attachment_type_set(plain, "mixed");
	attachment_basetype_set(plain, "multipart");
	
	data = gpgme_data_release_and_get_mem (dh_p, &length);
	dh_p = NULL;
	
	attachment_data_set(plain, data, length);
	
	gpgme_free(data);

err_dh_p:
	if (dh_p)
		gpgme_data_release(dh_p);
err_dh_e:
	gpgme_data_release(dh_e);
err_no_dh:
	return plain;
}

struct attachment *gpg_encrypt_sign(struct gpg *gpg, struct attachment *a, char *to)
{
	gpgme_error_t err;
	gpgme_data_t plain = NULL;
	gpgme_data_t cipher = NULL;
	struct attachment *encrypted = NULL;
	void *data;
	size_t length;
	gpgme_key_t recipients[2] = { NULL, NULL };
	
	err = gpgme_op_keylist_start (gpg->ctx, to, 0);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating data struct\n");
		goto err;
	}
	
	do {
		err = gpgme_op_keylist_next (gpg->ctx, recipients);
	} while (err == GPG_ERR_EOF);

	err = gpgme_data_new_from_mem(&plain, a->data, a->length, 0);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating encrypted data struct\n");
		goto err_data;
	}

	err = gpgme_data_new(&cipher);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error creating plain data struct\n");
		goto err_data;
	}
	
	err = gpgme_op_encrypt_sign(gpg->ctx, recipients,
	   GPGME_ENCRYPT_ALWAYS_TRUST, plain, cipher);
	if (err != GPG_ERR_NO_ERROR) {
		printf("Error encrypting and signing\n");
		goto err_data;
	}

	encrypted = attachment_alloc();
	if (!encrypted)
		goto err_data;

	attachment_type_set(encrypted, "octet-stream");
	attachment_basetype_set(encrypted, "application");
	
	data = gpgme_data_release_and_get_mem (cipher, &length);
	cipher = NULL;
	
	attachment_data_set(encrypted, data, length);
	
	gpgme_free(data);

	gpgme_key_release(recipients[0]);

err_data:
	if (cipher)
		gpgme_data_release(cipher);
	if (plain)
		gpgme_data_release(plain);
	
err:
	return encrypted;
}

#define PGPMIME_ID "Version: 1\n"

struct attachment *pgp_mime_id_get(void)
{
	struct attachment *a;
	
	a = attachment_alloc();
	attachment_data_set(a, PGPMIME_ID, strlen(PGPMIME_ID));
	attachment_type_set(a, "pgp-encrypted");
	attachment_basetype_set(a, "application");
	
	return a;
}
