*/
    QuickMSG
    Copyright (C) 2014  Jeroen Vreeken <jeroen@vreeken.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#include "mailbox.h"
#include "quickmsg.h"

#include <libetpan/libetpan.h>

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>


struct mailbox {
	char *server;
	unsigned short port;
	char *login;
	char *pass;
	
	struct mailimap *imap;
	bool loggedin;
};

static bool imap_error(int ret)
{
	if (ret == MAILIMAP_NO_ERROR)
		return false;
	if (ret == MAILIMAP_NO_ERROR_AUTHENTICATED)
		return false;
	if (ret == MAILIMAP_NO_ERROR_NON_AUTHENTICATED)
		return false;
	
	return true;
}

int mailbox_connect(struct mailbox *mailbox)
{
	unsigned short port_used;
	int r;

	if (!mailbox->port)
		port_used = 993;
	else
		port_used = mailbox->port;
	
	r = mailimap_ssl_connect(mailbox->imap, 
	    mailbox->server, port_used);

	if (imap_error(r)) {
		if (!mailbox->port)
			port_used = 143;
		else
			port_used = mailbox->port;
		
		r = mailimap_socket_connect(mailbox->imap, 
		    mailbox->server, port_used);
	}
	
	if (imap_error(r))
		return -1;


	r = mailimap_login(mailbox->imap, mailbox->login, mailbox->pass);
	if (imap_error(r))
		return -1;

	mailbox->loggedin = true;
		
	return 0;
}

int mailbox_disconnect(struct mailbox *mailbox)
{
	mailimap_logout(mailbox->imap);

	return 0;
}

int mailbox_chdir(struct mailbox *mailbox, char *dir)
{
	int r;
	
	r = mailimap_select(mailbox->imap, dir);

	return imap_error(r);
}


int mailbox_ls(struct mailbox *mailbox, uint32_t **uids, size_t *nr)
{
	struct mailimap_set *set;
	struct mailimap_fetch_type *fetch_type;
	struct mailimap_fetch_att *fetch_att;
	clist *fetch_result;
	clistiter *cur;
	int r, ret = 0;

	*nr = 0;
	*uids = NULL;
	
	set = mailimap_set_new_interval(1, 0);
	fetch_type = mailimap_fetch_type_new_fetch_att_list_empty();
	fetch_att = mailimap_fetch_att_new_uid();
	mailimap_fetch_type_new_fetch_att_list_add(fetch_type, fetch_att);
	
	r = mailimap_fetch(mailbox->imap, set, fetch_type, &fetch_result);
	if (imap_error(r))
		return -1;
	
	for (cur = clist_begin(fetch_result); cur != NULL; 
	    cur = clist_next(cur)) {
		struct mailimap_msg_att *msg_att;
		clistiter *cur2;
		
		msg_att = clist_content(cur);
		
		for (cur2 = clist_begin(msg_att->att_list); cur2 != NULL; 
		    cur2=clist_next(cur2)) {
			struct mailimap_msg_att_item *item;
			
			item = clist_content(cur2);
			
			if (item->att_type != MAILIMAP_MSG_ATT_ITEM_STATIC) {
				continue;
			}
			
			if (item->att_data.att_static->att_type != MAILIMAP_MSG_ATT_UID) {
				continue;
			}
			
			(*nr)++;
			*uids = realloc(*uids, sizeof(uint32_t) * (*nr));
			if (!*uids) {
				ret = -1;
				goto err_free;
			}
			(*uids)[(*nr) - 1] = item->att_data.att_static->att_data.att_uid;
		}
	}

err_free:
	mailimap_fetch_list_free(fetch_result);
	
	return ret;
}

int mailbox_get_id_is_deleted(struct mailbox *mailbox, uint32_t id, bool *deleted)
{
	struct mailimap_set *set;
	struct mailimap_fetch_type *fetch_type;
	struct mailimap_fetch_att *fetch_att;
	clist *fetch_result;
	clistiter *cur;
	int r, ret = 0;

	*deleted = false;
	
	set = mailimap_set_new_single(id);
	fetch_type = mailimap_fetch_type_new_fetch_att_list_empty();
	fetch_att = mailimap_fetch_att_new_flags();
	mailimap_fetch_type_new_fetch_att_list_add(fetch_type, fetch_att);
	
	r = mailimap_fetch(mailbox->imap, set, fetch_type, &fetch_result);
	if (imap_error(r)) {
		printf("error during fetch: %d\n", r);
		return -1;
	}
	
	for (cur = clist_begin(fetch_result); cur != NULL; 
	    cur = clist_next(cur)) {
		struct mailimap_msg_att *msg_att;
		clistiter *cur2;
		
		msg_att = clist_content(cur);
		
		for (cur2 = clist_begin(msg_att->att_list); cur2 != NULL; 
		    cur2=clist_next(cur2)) {
			struct mailimap_msg_att_item *item;
			clistiter *cur3;
			
			item = clist_content(cur2);
			
			if (item->att_type != MAILIMAP_MSG_ATT_ITEM_DYNAMIC) {
				continue;
			}
			if (item->att_data.att_dyn->att_list == NULL)
				continue;
			
			for (cur3 = clist_begin(item->att_data.att_dyn->att_list); 
			    cur3 != NULL; cur3 = clist_next(cur3)) {
				struct mailimap_flag_fetch *ff =
				    clist_content(cur3);
				
				if (ff->fl_flag->fl_type == MAILIMAP_FLAG_DELETED) {
					*deleted = true;
				}
			}
						    
		}
	}

//err_free:
	mailimap_fetch_list_free(fetch_result);
	
	return ret;
}

int mailbox_get_id_headers(struct mailbox *mailbox, uint32_t id, char **header, size_t *length)
{
	struct mailimap_set *set;
	struct mailimap_fetch_type *fetch_type;
	struct mailimap_fetch_att *fetch_att;
	clist *fetch_result;
	clistiter *cur;
	int r, ret = 0;

	*length = 0;
	*header = NULL;
	
	set = mailimap_set_new_single(id);
	fetch_type = mailimap_fetch_type_new_fetch_att_list_empty();
	fetch_att = mailimap_fetch_att_new_rfc822_header();
	mailimap_fetch_type_new_fetch_att_list_add(fetch_type, fetch_att);
	
	r = mailimap_fetch(mailbox->imap, set, fetch_type, &fetch_result);
	if (imap_error(r)) {
		printf("error during fetch: %d\n", r);
		return -1;
	}
	
	for (cur = clist_begin(fetch_result); cur != NULL; 
	    cur = clist_next(cur)) {
		struct mailimap_msg_att *msg_att;
		clistiter *cur2;
		
		printf("iter\n");
		msg_att = clist_content(cur);
		
		for (cur2 = clist_begin(msg_att->att_list); cur2 != NULL; 
		    cur2=clist_next(cur2)) {
			struct mailimap_msg_att_item *item;
			
			item = clist_content(cur2);
			
			if (item->att_type != MAILIMAP_MSG_ATT_ITEM_STATIC) {
				continue;
			}
			
			if (item->att_data.att_static->att_type != MAILIMAP_MSG_ATT_RFC822_HEADER) {
				continue;
			}
			
			*length = item->att_data.att_static->att_data.att_rfc822_header.att_length;
						    
			*header = malloc((*length) + 1);
			if (!*header)
				goto err_free;
		
			memcpy(*header, 
			    item->att_data.att_static->att_data.att_rfc822_header.att_content,
			    *length);
			(*header)[*length] = 0;
		}
	}

err_free:
	mailimap_fetch_list_free(fetch_result);
	
	return ret;
}

int mailbox_get_id_is_quickmsg(struct mailbox *mailbox, uint32_t id, bool *quickmsg)
{
	char *header = NULL;
	size_t length;
	char *ptr;

	int r = mailbox_get_id_headers(mailbox, id, &header, &length);
	
	for (ptr = header; *ptr; ptr++) {
		if (*ptr == '\n' || *ptr == '\r') {
			if (!strncasecmp(ptr+1, QUICKMSG_MAIL_HEADER,
			    strlen(QUICKMSG_MAIL_HEADER))) {
				*quickmsg = true;
			}
		}
	}
		
	free(header);
	return r;
}

static char * get_msg_att_msg_content(struct mailimap_msg_att * msg_att, size_t * p_msg_size)
{
	clistiter * cur;

	/* iterate on each result of one given message */
	for(cur = clist_begin(msg_att->att_list) ; cur != NULL ; cur = clist_next(cur)) {
		struct mailimap_msg_att_item * item;

		item = clist_content(cur);
		if (item->att_type != MAILIMAP_MSG_ATT_ITEM_STATIC) {
			continue;
		}

		if (item->att_data.att_static->att_type != MAILIMAP_MSG_ATT_BODY_SECTION) {
			continue;
		}

		*p_msg_size = item->att_data.att_static->att_data.att_body_section->sec_length;
		return item->att_data.att_static->att_data.att_body_section->sec_body_part;
	}

	return NULL;
}

int mailbox_get_id_body(struct mailbox *mailbox, uint32_t id, char **body, size_t *length)
{
	struct mailimap_set *set;
	struct mailimap_section *section;
	struct mailimap_fetch_type *fetch_type;
	struct mailimap_fetch_att *fetch_att;
	clist *fetch_result;
	clistiter *cur;
	int r, ret = 0;

	*length = 0;
	*body = NULL;
	
	set = mailimap_set_new_single(id);
	fetch_type = mailimap_fetch_type_new_fetch_att_list_empty();
	section = mailimap_section_new(NULL);
	fetch_att = mailimap_fetch_att_new_body_peek_section(section);
	mailimap_fetch_type_new_fetch_att_list_add(fetch_type, fetch_att);
	
	r = mailimap_fetch(mailbox->imap, set, fetch_type, &fetch_result);
	if (imap_error(r)) {
		printf("error during fetch: %d\n", r);
		return -1;
	}
	
	for(cur = clist_begin(fetch_result) ; cur != NULL ; cur = clist_next(cur)) {
		struct mailimap_msg_att * msg_att;
		size_t msg_size;
		char *msg_content;

		msg_att = clist_content(cur);
		msg_content = get_msg_att_msg_content(msg_att, &msg_size);
		if (msg_content == NULL) {
			continue;
		}

		*length = msg_size;

		*body = malloc((*length) + 1);
		if (!*body)
			goto err_free;
		
		memcpy(*body, msg_content, *length);
		(*body)[*length] = 0;
	}
	
err_free:
	mailimap_fetch_list_free(fetch_result);
	
	return ret;
}

int mailbox_delete_id(struct mailbox *mailbox, uint32_t id)
{
	struct mailimap_set *set;
	struct mailimap_store_att_flags *flags;
	struct mailimap_flag *flag;
	struct mailimap_flag_list *flag_list;
	clist *fl_list;
	
	fl_list = clist_new();

	flag = mailimap_flag_new(MAILIMAP_FLAG_DELETED, NULL, NULL);
	clist_prepend(fl_list, flag);
	flag_list = mailimap_flag_list_new(fl_list);
	
	set = mailimap_set_new_single(id);
	flags = mailimap_store_att_flags_new(1, 0, flag_list);

	mailimap_store(mailbox->imap, set, flags);

	mailimap_store_att_flags_free(flags);

	return 0;
}

int mailbox_mkdir(struct mailbox *mailbox, char *dir)
{
	int r;
	
	r = mailimap_create(mailbox->imap, dir);
	
	if (imap_error(r))
		return -1;
	
	return 0;
}

struct mailbox *mailbox_create(char *server, unsigned short port, char *login, char *pass)
{
	struct mailbox *mailbox;
	
	mailbox = calloc(sizeof(struct mailbox), 1);
	
	mailbox->imap = mailimap_new(0, NULL);
	if (!mailbox->imap)
		goto err_mailbox;
	
	mailbox->server = strdup(server);
	mailbox->port = port;
	mailbox->login = strdup(login);
	mailbox->pass = strdup(pass);

	if (!mailbox->server || !mailbox->login || !mailbox->pass)
		goto err_mem;
	
	return mailbox;

err_mem:
	free(mailbox->server);
	free(mailbox->login);
	free(mailbox->pass);
	mailimap_free(mailbox->imap);
err_mailbox:
	free(mailbox);
	return NULL;
}

void mailbox_destroy(struct mailbox *mailbox)
{
	if (mailbox->imap) {
		mailimap_logout(mailbox->imap);
		mailimap_free(mailbox->imap);
	}
	free(mailbox->server);
	free(mailbox->login);
	free(mailbox->pass);
	free(mailbox);
}
